@@ -2,6 +2,7 @@ import numpy as np
 np.seterr(all='raise')
 from typing import Dict, Union
 
+from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLActionPreconditionNotSatisfiedError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidActionError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
@@ -11,13 +12,12 @@ from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError
 from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLValueOutOfRangeError
 
-from pyRDDLGym.Core.Compiler.RDDLDecompiler import RDDLDecompiler
 from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
 from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
 from pyRDDLGym.Core.Compiler.RDDLObjectsTracer import RDDLObjectsTracer
 from pyRDDLGym.Core.Compiler.RDDLValueInitializer import RDDLValueInitializer
 from pyRDDLGym.Core.Debug.Logger import Logger
-from pyRDDLGym.Core.Parser.expr import Expression, Value
+from pyRDDLGym.Core.Parser.expr import Value
 
 Args = Dict[str, Value]
 
@@ -36,38 +36,11 @@ class RDDLSimulator:
         :param logger: to log information about compilation to file
         '''
         self.rddl = rddl
+        self.allow_synchronous_state = allow_synchronous_state
         self.rng = rng
         self.logger = logger
         
-        # compile initial values
-        if self.logger is not None:
-            self.logger.clear()
-        initializer = RDDLValueInitializer(rddl, logger=self.logger)
-        self.init_values = initializer.initialize()
-        
-        # compute dependency graph for CPFs and sort them by evaluation order
-        sorter = RDDLLevelAnalysis(rddl, allow_synchronous_state, logger=self.logger)
-        levels = sorter.compute_levels()      
-        self.cpfs = []  
-        for cpfs in levels.values():
-            for cpf in cpfs:
-                _, expr = rddl.cpfs[cpf]
-                prange = rddl.variable_ranges[cpf]
-                dtype = RDDLValueInitializer.NUMPY_TYPES.get(
-                    prange, RDDLValueInitializer.INT)
-                self.cpfs.append((cpf, expr, dtype))
-                
-        # trace expressions to cache information to be used later
-        tracer = RDDLObjectsTracer(rddl, logger=self.logger)
-        self.traced = tracer.trace()
-        
-        # initialize all fluent and non-fluent values        
-        self.subs = self.init_values.copy()
-        self.state = None  
-        self.noop_actions = {var: values
-                             for (var, values) in self.init_values.items()
-                             if rddl.variable_types[var] == 'action-fluent'}
-        self._pomdp = bool(rddl.observ)
+        self._compile()
         
         # basic operations
         self.ARITHMETIC_OPS = {
@@ -130,11 +103,46 @@ class RDDLSimulator:
             'min': np.minimum,
             'max': np.maximum,
             'pow': np.power,
-            'log': lambda x, y: np.log(x) / np.log(y)
+            'log': lambda x, y: np.log(x) / np.log(y),
+            'hypot': np.hypot
         }
         self.CONTROL_OPS = {'if': np.where,
                             'switch': np.select}
     
+    def _compile(self):
+        rddl = self.rddl
+        
+        # compile initial values
+        if self.logger is not None:
+            self.logger.clear()
+        initializer = RDDLValueInitializer(rddl, logger=self.logger)
+        self.init_values = initializer.initialize()
+        
+        # compute dependency graph for CPFs and sort them by evaluation order
+        sorter = RDDLLevelAnalysis(rddl, self.allow_synchronous_state, 
+                                   logger=self.logger)
+        levels = sorter.compute_levels()      
+        self.cpfs = []  
+        for cpfs in levels.values():
+            for cpf in cpfs:
+                _, expr = rddl.cpfs[cpf]
+                prange = rddl.variable_ranges[cpf]
+                dtype = RDDLValueInitializer.NUMPY_TYPES.get(
+                    prange, RDDLValueInitializer.INT)
+                self.cpfs.append((cpf, expr, dtype))
+                
+        # trace expressions to cache information to be used later
+        tracer = RDDLObjectsTracer(rddl, logger=self.logger)
+        self.traced = tracer.trace()
+        
+        # initialize all fluent and non-fluent values        
+        self.subs = self.init_values.copy()
+        self.state = None  
+        self.noop_actions = {var: values
+                             for (var, values) in self.init_values.items()
+                             if rddl.variable_types[var] == 'action-fluent'}
+        self._pomdp = bool(rddl.observ)
+        
     @property
     def states(self) -> Args:
         return self.state.copy()
@@ -147,28 +155,18 @@ class RDDLSimulator:
     # error checks
     # ===========================================================================
     
-    @staticmethod
-    def _print_stack_trace(expr):
-        if isinstance(expr, Expression):
-            trace = RDDLDecompiler().decompile_expr(expr)
-        else:
-            trace = str(expr)
-        return '>> ' + trace
-    
     @staticmethod
     def _check_type(value, valid, msg, expr, arg=None):
         if not np.can_cast(value, valid):
             dtype = getattr(value, 'dtype', type(value))
             if arg is None:
                 raise RDDLTypeError(
-                    f'{msg} must evaluate to {valid}, '
-                    f'got {value} of type {dtype}.\n' + 
-                    RDDLSimulator._print_stack_trace(expr))
+                    f'{msg} must evaluate to <{valid}>, '
+                    f'got {value} of type <{dtype}>.\n' + print_stack_trace(expr))
             else:
                 raise RDDLTypeError(
-                    f'Argument {arg} of {msg} must evaluate to {valid}, '
-                    f'got {value} of type {dtype}.\n' + 
-                    RDDLSimulator._print_stack_trace(expr))
+                    f'Argument {arg} of {msg} must evaluate to <{valid}>, '
+                    f'got {value} of type <{dtype}>.\n' + print_stack_trace(expr))
     
     @staticmethod
     def _check_types(value, valid, msg, expr):
@@ -178,8 +176,7 @@ class RDDLSimulator:
         dtype = getattr(value, 'dtype', type(value))
         raise RDDLTypeError(
             f'{msg} must evaluate to one of {valid}, '
-            f'got {value} of type {dtype}.\n' + 
-            RDDLSimulator._print_stack_trace(expr))
+            f'got {value} of type <{dtype}>.\n' + print_stack_trace(expr))
     
     @staticmethod
     def _check_op(op, valid, msg, expr):
@@ -187,27 +184,26 @@ class RDDLSimulator:
         if numpy_op is None:
             raise RDDLNotImplementedError(
                 f'{msg} operator {op} is not supported: '
-                f'must be in {set(valid.keys())}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                f'must be in {set(valid.keys())}.\n' + print_stack_trace(expr))
         return numpy_op
         
     @staticmethod
     def _check_arity(args, required, msg, expr):
         if len(args) != required:
             raise RDDLInvalidNumberOfArgumentsError(
-                f'{msg} requires {required} arguments, got {len(args)}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                f'{msg} requires {required} argument(s), got {len(args)}.\n' + 
+                print_stack_trace(expr))
     
     @staticmethod
     def _check_positive(value, strict, msg, expr):
         if strict and not np.all(value > 0):
             raise RDDLValueOutOfRangeError(
                 f'{msg} must be positive, got {value}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
         elif not strict and not np.all(value >= 0):
             raise RDDLValueOutOfRangeError(
                 f'{msg} must be non-negative, got {value}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
     
     @staticmethod
     def _check_bounds(lb, ub, msg, expr):
@@ -215,30 +211,29 @@ class RDDLSimulator:
             raise RDDLValueOutOfRangeError(
                 f'Bounds of {msg} are invalid:' 
                 f'max value {ub} must be >= min value {lb}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
             
     @staticmethod
     def _check_range(value, lb, ub, msg, expr):
         if not np.all(np.logical_and(value >= lb, value <= ub)):
             raise RDDLValueOutOfRangeError(
                 f'{msg} must be in the range [{lb}, {ub}], got {value}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
     
     # ===========================================================================
     # main sampling routines
     # ===========================================================================
     
     def _process_actions(self, actions):
+        rddl = self.rddl
         new_actions = {action: np.copy(value) 
                        for (action, value) in self.noop_actions.items()}
         
         # override new_actions with any new actions
-        rddl = self.rddl
         for (action, value) in actions.items(): 
             
-            # enum literals are converted to their canonical int indices
-            if value in rddl.enum_literals:
-                value = rddl.index_of_object[value]
+            # objects are converted to their canonical indices
+            value = rddl.index_of_object.get(value, value)
             
             # parse action string and assign to the correct coordinates
             if action in new_actions:
@@ -263,7 +258,7 @@ class RDDLSimulator:
             if not bool(sample):
                 raise RDDLStateInvariantNotSatisfiedError(
                     f'Invariant {i + 1} is not satisfied.\n' + 
-                    RDDLSimulator._print_stack_trace(invariant))
+                    print_stack_trace(invariant))
     
     def check_action_preconditions(self, actions: Args) -> None:
         '''Throws an exception if the action preconditions are not satisfied.'''        
@@ -276,7 +271,7 @@ class RDDLSimulator:
             if not bool(sample):
                 raise RDDLActionPreconditionNotSatisfiedError(
                     f'Precondition {i + 1} is not satisfied.\n' + 
-                    RDDLSimulator._print_stack_trace(precond))
+                    print_stack_trace(precond))
     
     def check_terminal_states(self) -> bool:
         '''Return True if a terminal state has been reached.'''
@@ -293,10 +288,10 @@ class RDDLSimulator:
     
     def reset(self) -> Union[Dict[str, None], Args]:
         '''Resets the state variables to their initial values.'''
+        rddl = self.rddl
         subs = self.subs = self.init_values.copy()
         
         # update state
-        rddl = self.rddl
         self.state = {}
         for state in rddl.states:
             self.state.update(rddl.ground_values(state, subs[state]))
@@ -315,12 +310,12 @@ class RDDLSimulator:
         
         :param actions: a dict mapping current action fluent to their values
         '''
+        rddl = self.rddl
         actions = self._process_actions(actions)
         subs = self.subs
         subs.update(actions)
         
         # evaluate CPFs in topological order
-        rddl = self.rddl
         for (cpf, expr, dtype) in self.cpfs:
             sample = self._sample(expr, subs)
             RDDLSimulator._check_type(sample, dtype, cpf, expr)
@@ -377,7 +372,7 @@ class RDDLSimulator:
         else:
             raise RDDLNotImplementedError(
                 f'Internal error: expression type {etype} is not supported.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
                 
     # ===========================================================================
     # leaves
@@ -389,27 +384,23 @@ class RDDLSimulator:
     def _sample_pvar(self, expr, subs):
         var, args = expr.args
         
-        # free variable (e.g., ?x) treated as integer array (0, 1, 2...)
-        if self.rddl.is_free_variable(var):
-            return self.traced.cached_sim_info(expr)
-        
-        # literal of enumerated type is treated as integer
-        elif not args and self.rddl.is_literal(var):
-            return self.traced.cached_sim_info(expr)
+        # free variable (e.g., ?x) and object converted to canonical index
+        is_value, cached_info = self.traced.cached_sim_info(expr)
+        if is_value:
+            return cached_info
         
         # extract variable value
         sample = subs.get(var, None)
         if sample is None:
             raise RDDLUndefinedVariableError(
                 f'Variable <{var}> is referenced before assignment.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
         
         # lifted domain must slice and/or reshape value tensor
-        shape_info = self.traced.cached_sim_info(expr)
-        if shape_info is not None:
-            slices, axis, shape, op_code, op_args = shape_info
+        if cached_info is not None:
+            slices, axis, shape, op_code, op_args = cached_info
             if slices: 
-                if op_code == -1:
+                if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.NESTED_SLICE:
                     slices = tuple(
                         (self._sample(arg, subs) if _slice is None else _slice)
                         for (arg, _slice) in zip(args, slices)
@@ -418,9 +409,9 @@ class RDDLSimulator:
             if axis:
                 sample = np.expand_dims(sample, axis=axis)
                 sample = np.broadcast_to(sample, shape=shape)
-            if op_code == 0:
+            if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.EINSUM:
                 sample = np.einsum(sample, *op_args)
-            elif op_code == 1:
+            elif op_code == RDDLObjectsTracer.NUMPY_OP_CODE.TRANSPOSE:
                 sample = np.transpose(sample, axes=op_args)
         return sample
     
@@ -455,7 +446,7 @@ class RDDLSimulator:
                     raise ArithmeticError(
                         f'Cannot evaluate arithmetic operation {op} '
                         f'at {sample_lhs} and {sample_rhs}.\n' + 
-                        RDDLSimulator._print_stack_trace(expr))
+                        print_stack_trace(expr))
         
         # for a grounded domain can short-circuit * and +
         elif n > 0 and not self.traced.cached_objects_in_scope(expr):
@@ -466,8 +457,7 @@ class RDDLSimulator:
         
         raise RDDLInvalidNumberOfArgumentsError(
             f'Arithmetic operator {op} does not have the required '
-            f'number of arguments.\n' + 
-            RDDLSimulator._print_stack_trace(expr))
+            f'number of arguments.\n' + print_stack_trace(expr))
     
     def _sample_product(self, lhs, rhs, subs):
         
@@ -556,8 +546,7 @@ class RDDLSimulator:
             
         raise RDDLInvalidNumberOfArgumentsError(
             f'Logical operator {op} does not have the required '
-            f'number of arguments.\n' + 
-            RDDLSimulator._print_stack_trace(expr))
+            f'number of arguments.\n' + print_stack_trace(expr))
     
     def _sample_and_or(self, lhs, rhs, op, expr, subs):
         
@@ -646,7 +635,7 @@ class RDDLSimulator:
             except:
                 raise ArithmeticError(
                     f'Cannot evaluate unary function {name} at {sample}.\n' + 
-                    RDDLSimulator._print_stack_trace(expr))
+                    print_stack_trace(expr))
         
         # binary function
         binary_op = self.BINARY.get(name, None)
@@ -660,12 +649,10 @@ class RDDLSimulator:
             except:
                 raise ArithmeticError(
                     f'Cannot evaluate binary function {name} at '
-                    f'{sample_lhs} and {sample_rhs}.\n' + 
-                    RDDLSimulator._print_stack_trace(expr))
+                    f'{sample_lhs} and {sample_rhs}.\n' + print_stack_trace(expr))
         
         raise RDDLNotImplementedError(
-            f'Function {name} is not supported.\n' + 
-            RDDLSimulator._print_stack_trace(expr))
+            f'Function {name} is not supported.\n' + print_stack_trace(expr))
     
     # ===========================================================================
     # control flow
@@ -777,6 +764,8 @@ class RDDLSimulator:
             return self._sample_gompertz(expr, subs)
         elif name == 'ChiSquare':
             return self._sample_chisquare(expr, subs)
+        elif name == 'Kumaraswamy':
+            return self._sample_kumaraswamy(expr, subs)
         elif name == 'Discrete':
             return self._sample_discrete(expr, subs, unnorm=False)
         elif name == 'UnnormDiscrete':
@@ -788,7 +777,7 @@ class RDDLSimulator:
         else:
             raise RDDLNotImplementedError(
                 f'Distribution {name} is not supported.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
 
     def _sample_kron_delta(self, expr, subs):
         args = expr.args
@@ -818,7 +807,7 @@ class RDDLSimulator:
         sample_lb = self._sample(lb, subs)
         sample_ub = self._sample(ub, subs)
         RDDLSimulator._check_bounds(sample_lb, sample_ub, 'Uniform', expr)
-        return self.rng.uniform(sample_lb, sample_ub)      
+        return self.rng.uniform(low=sample_lb, high=sample_ub)      
     
     def _sample_bernoulli(self, expr, subs):
         args = expr.args
@@ -839,7 +828,7 @@ class RDDLSimulator:
         sample_var = self._sample(var, subs)
         RDDLSimulator._check_positive(sample_var, False, 'Normal variance', expr)  
         sample_std = np.sqrt(sample_var)
-        return self.rng.normal(sample_mean, sample_std)
+        return self.rng.normal(loc=sample_mean, scale=sample_std)
     
     def _sample_poisson(self, expr, subs):
         args = expr.args
@@ -848,7 +837,7 @@ class RDDLSimulator:
         rate, = args
         sample_rate = self._sample(rate, subs)
         RDDLSimulator._check_positive(sample_rate, False, 'Poisson rate', expr)        
-        return self.rng.poisson(sample_rate)
+        return self.rng.poisson(lam=sample_rate)
     
     def _sample_exponential(self, expr, subs):
         args = expr.args
@@ -857,7 +846,7 @@ class RDDLSimulator:
         scale, = expr.args
         sample_scale = self._sample(scale, subs)
         RDDLSimulator._check_positive(sample_scale, True, 'Exponential rate', expr)
-        return self.rng.exponential(sample_scale)
+        return self.rng.exponential(scale=sample_scale)
     
     def _sample_weibull(self, expr, subs):
         args = expr.args
@@ -868,7 +857,7 @@ class RDDLSimulator:
         sample_scale = self._sample(scale, subs)
         RDDLSimulator._check_positive(sample_shape, True, 'Weibull shape', expr)
         RDDLSimulator._check_positive(sample_scale, True, 'Weibull scale', expr)
-        return sample_scale * self.rng.weibull(sample_shape)
+        return sample_scale * self.rng.weibull(a=sample_shape)
     
     def _sample_gamma(self, expr, subs):
         args = expr.args
@@ -879,7 +868,7 @@ class RDDLSimulator:
         sample_scale = self._sample(scale, subs)
         RDDLSimulator._check_positive(sample_shape, True, 'Gamma shape', expr)            
         RDDLSimulator._check_positive(sample_scale, True, 'Gamma scale', expr)        
-        return self.rng.gamma(sample_shape, sample_scale)
+        return self.rng.gamma(shape=sample_shape, scale=sample_scale)
     
     def _sample_binomial(self, expr, subs):
         args = expr.args
@@ -891,7 +880,7 @@ class RDDLSimulator:
         RDDLSimulator._check_type(sample_count, RDDLValueInitializer.INT, 'Binomial count', expr)
         RDDLSimulator._check_positive(sample_count, False, 'Binomial count', expr)
         RDDLSimulator._check_range(sample_pr, 0, 1, 'Binomial p', expr)
-        return self.rng.binomial(sample_count, sample_pr)
+        return self.rng.binomial(n=sample_count, p=sample_pr)
     
     def _sample_negative_binomial(self, expr, subs):
         args = expr.args
@@ -902,7 +891,7 @@ class RDDLSimulator:
         sample_pr = self._sample(pr, subs)
         RDDLSimulator._check_positive(sample_count, True, 'NegativeBinomial r', expr)
         RDDLSimulator._check_range(sample_pr, 0, 1, 'NegativeBinomial p', expr)        
-        return self.rng.negative_binomial(sample_count, sample_pr)
+        return self.rng.negative_binomial(n=sample_count, p=sample_pr)
     
     def _sample_beta(self, expr, subs):
         args = expr.args
@@ -913,7 +902,7 @@ class RDDLSimulator:
         sample_rate = self._sample(rate, subs)
         RDDLSimulator._check_positive(sample_shape, True, 'Beta shape', expr)
         RDDLSimulator._check_positive(sample_rate, True, 'Beta rate', expr)        
-        return self.rng.beta(sample_shape, sample_rate)
+        return self.rng.beta(a=sample_shape, b=sample_rate)
 
     def _sample_geometric(self, expr, subs):
         args = expr.args
@@ -922,7 +911,7 @@ class RDDLSimulator:
         pr, = args
         sample_pr = self._sample(pr, subs)
         RDDLSimulator._check_range(sample_pr, 0, 1, 'Geometric p', expr)        
-        return self.rng.geometric(sample_pr)
+        return self.rng.geometric(p=sample_pr)
     
     def _sample_pareto(self, expr, subs):
         args = expr.args
@@ -933,7 +922,7 @@ class RDDLSimulator:
         sample_scale = self._sample(scale, subs)
         RDDLSimulator._check_positive(sample_shape, True, 'Pareto shape', expr)        
         RDDLSimulator._check_positive(sample_scale, True, 'Pareto scale', expr)        
-        return sample_scale * self.rng.pareto(sample_shape)
+        return sample_scale * self.rng.pareto(a=sample_shape)
     
     def _sample_student(self, expr, subs):
         args = expr.args
@@ -942,7 +931,7 @@ class RDDLSimulator:
         df, = args
         sample_df = self._sample(df, subs)
         RDDLSimulator._check_positive(sample_df, True, 'Student df', expr)            
-        return self.rng.standard_t(sample_df)
+        return self.rng.standard_t(df=sample_df)
 
     def _sample_gumbel(self, expr, subs):
         args = expr.args
@@ -952,7 +941,7 @@ class RDDLSimulator:
         sample_mean = self._sample(mean, subs)
         sample_scale = self._sample(scale, subs)
         RDDLSimulator._check_positive(sample_scale, True, 'Gumbel scale', expr)
-        return self.rng.gumbel(sample_mean, sample_scale)
+        return self.rng.gumbel(loc=sample_mean, scale=sample_scale)
     
     def _sample_laplace(self, expr, subs):
         args = expr.args
@@ -962,7 +951,7 @@ class RDDLSimulator:
         sample_mean = self._sample(mean, subs)
         sample_scale = self._sample(scale, subs)
         RDDLSimulator._check_positive(sample_scale, True, 'Laplace scale', expr)
-        return self.rng.laplace(sample_mean, sample_scale)
+        return self.rng.laplace(loc=sample_mean, scale=sample_scale)
     
     def _sample_cauchy(self, expr, subs):
         args = expr.args
@@ -998,6 +987,19 @@ class RDDLSimulator:
         RDDLSimulator._check_positive(sample_df, True, 'ChiSquare df', expr)
         return self.rng.chisquare(df=sample_df)
     
+    def _sample_kumaraswamy(self, expr, subs):
+        args = expr.args
+        RDDLSimulator._check_arity(args, 2, 'Kumaraswamy', expr)
+        
+        a, b = args
+        sample_a = self._sample(a, subs)
+        sample_b = self._sample(b, subs)
+        RDDLSimulator._check_positive(sample_a, True, 'Kumaraswamy a', expr)
+        RDDLSimulator._check_positive(sample_b, True, 'Kumaraswamy b', expr)
+        size = sample_a.shape if self.traced.cached_objects_in_scope(expr) else None
+        U = self.rng.uniform(size=size)
+        return (1.0 - U ** (1.0 / sample_b)) ** (1.0 / sample_a)
+    
     # ===========================================================================
     # random variables with enum support
     # ===========================================================================
@@ -1014,7 +1016,7 @@ class RDDLSimulator:
         if not np.allclose(cdf[..., -1], 1.0):
             raise RDDLValueOutOfRangeError(
                 f'Discrete probabilities must sum to 1, got {cdf[..., -1]}.\n' + 
-                RDDLSimulator._print_stack_trace(expr))     
+                print_stack_trace(expr))     
         
         # use inverse CDF sampling                  
         U = self.rng.random(size=cdf.shape[:-1] + (1,))
@@ -1049,7 +1051,7 @@ class RDDLSimulator:
         else:
             raise RDDLNotImplementedError(
                 f'Multivariate distribution {name} is not supported.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
     
     def _sample_multivariate_normal(self, expr, subs):
         _, args = expr.args
@@ -1113,38 +1115,31 @@ class RDDLSimulator:
         sample = np.moveaxis(sample, source=-1, destination=index)
         return sample
     
-    def _discrete_sample_to_mask(self, sample, num_categories):
-        categories = np.arange(num_categories)
-        categories = categories[(...,) + (np.newaxis,) * len(sample.shape)]
-        sample = sample[np.newaxis, ...]
-        return (sample == categories)
-        
     def _sample_multinomial(self, expr, subs):
         _, args = expr.args
         RDDLSimulator._check_arity(args, 2, 'Multinomial', expr)
         
-        prob, trials = args
-        sample_prob = self._sample(prob, subs)
-        sample_trials = self._sample(trials, subs)        
+        trials, prob = args
+        sample_trials = self._sample(trials, subs) 
+        sample_prob = self._sample(prob, subs)       
         RDDLSimulator._check_type(sample_trials, RDDLValueInitializer.INT, 'Multinomial trials', expr)
         RDDLSimulator._check_positive(sample_trials, False, 'Multinomial trials', expr)
+        RDDLSimulator._check_positive(sample_prob, False, 'Discrete probabilities', expr)
         
-        # prepend a trial axis to sample_prob
-        num_trials = int(sample_trials.flat[0])
-        shape = (num_trials,) + sample_prob.shape
-        sample_prob = sample_prob[np.newaxis, ...]
-        sample_prob = np.broadcast_to(sample_prob, shape=shape)
+        # check valid PMF
+        cum_prob = np.sum(sample_prob, axis=-1)
+        if not np.allclose(cum_prob, 1.0):
+            raise RDDLValueOutOfRangeError(
+                f'Multinomial probabilities must sum to 1, got {cum_prob}.\n' + 
+                print_stack_trace(expr))    
+            
+        # sample from the multinomial
+        sample = self.rng.multinomial(n=sample_trials, pvals=sample_prob)
         
-        # sample from the discrete distribution, convert to mask and sum trials
-        sample_discrete = self._sample_discrete_helper(sample_prob, False, expr)
-        num_categories = sample_prob.shape[-1]
-        masked = self._discrete_sample_to_mask(sample_discrete, num_categories)
-        sample = np.sum(masked, axis=1)
-        
-        # since the sampling is done in the first dimension we need to move it
+        # since the sampling is done in the last dimension we need to move it
         # to match the order of the CPF variables
         index, = self.traced.cached_sim_info(expr)
-        sample = np.moveaxis(sample, source=0, destination=index)
+        sample = np.moveaxis(sample, source=-1, destination=index)
         return sample
         
     # ===========================================================================
@@ -1156,21 +1151,24 @@ class RDDLSimulator:
         if op == 'det':
             return self._sample_matrix_det(expr, subs)
         elif op == 'inverse':
-            return self._sample_matrix_inv(expr, subs)
+            return self._sample_matrix_inv(expr, subs, pseudo=False)
+        elif op == 'pinverse':
+            return self._sample_matrix_inv(expr, subs, pseudo=True)
         else:
             raise RDDLNotImplementedError(
                 f'Matrix operator {op} is not supported.\n' + 
-                RDDLSimulator._print_stack_trace(expr))
+                print_stack_trace(expr))
     
     def _sample_matrix_det(self, expr, subs):
         * _, arg = expr.args
         sample_arg = self._sample(arg, subs)
         return np.linalg.det(sample_arg)
     
-    def _sample_matrix_inv(self, expr, subs):
+    def _sample_matrix_inv(self, expr, subs, pseudo):
         _, arg = expr.args
         sample_arg = self._sample(arg, subs)
-        sample = np.linalg.inv(sample_arg)
+        op = np.linalg.pinv if pseudo else np.linalg.inv
+        sample = op(sample_arg)
         
         # matrix dimensions are last two axes, move them to the correct position
         indices = self.traced.cached_sim_info(expr)
